from skyciv.classes.model.components.area_loads.area_load import AreaLoad
from skyciv.classes.model.components.distributed_loads.distributed_load import DistributedLoad
from skyciv.classes.model.components.load_combinations.load_combination import LoadCombination
from skyciv.classes.model.components.materials.material import Material
from skyciv.classes.model.components.members.member import Member
from skyciv.classes.model.components.meshed_plates.meshed_plate import MeshedPlate
from skyciv.classes.model.components.moments.moment import Moment
from skyciv.classes.model.components.nodes.node import Node
from skyciv.classes.model.components.plates.plate import Plate
from skyciv.classes.model.components.point_loads.point_load import PointLoad
from skyciv.classes.model.components.pressures.pressure import Pressure
from skyciv.classes.model.components.sections.section import Section
from skyciv.classes.model.components.self_weights.self_weight import SelfWeight
from skyciv.classes.model.components.settlements.settlement import Settlement
from skyciv.classes.model.components.supports.support import Support
from skyciv.classes.model.model import Model
from skyciv.utils.helpers import keys, keyvals

key_class_map = {
    "nodes": Node,
    "members": Member,
    "plates": Plate,
    "meshed_plates": MeshedPlate,
    "sections": Section,
    "materials": Material,
    "supports": Support,
    "settlements": Settlement,
    "point_loads": PointLoad,
    "moments": Moment,
    "distributed_loads": DistributedLoad,
    "pressures": Pressure,
    "area_loads": AreaLoad,
    "self_weight": SelfWeight,
    "load_combinations": LoadCombination,
}


def set_attribs(_from, to):
    for collection, collection_values in keyvals(_from):
        if (type(collection_values) is dict or hasattr(collection_values, "__dict__")):
            # Check if single item should be created
            if collection in keys(key_class_map):
                # print(collection)
                for id in keys(collection_values):
                    # print(id)
                    item = key_class_map[collection]()
                    # print(item)
                    setattr(to[collection], id, item)
            # Dig deeper
            if collection_values != None:
                if collection == "sections":
                    # Special treatment for sections
                    for id, sec_data in keyvals(collection_values):
                        if type(sec_data) == list:
                            setattr(to[collection][id],
                                    "load_section", sec_data)
                        elif sec_data is None:
                            setattr(to[collection], id, None)
                        else:
                            setattr(to[collection][id], "load_section",
                                    sec_data["aux"]["polygons"][0]["library_selections"])

                        setattr(to[collection][id], "material_id",
                                sec_data["material_id"])
                else:
                    set_attribs(_from[collection], to[collection])

        else:
            # Set the value
            print(collection_values)
            setattr(to, collection, collection_values)


def model_adapter(jsonModel, model):
    set_attribs(jsonModel, model)



from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent
from skyciv.classes.model.components.distributed_loads.distributed_load import DistributedLoad
from skyciv.utils.helpers import next_object_key


class DistributedLoads(ModelCollectionComponent):
    """Creates an instance of the SkyCiv DistributedLoads class.
    """

    def add(
        self,
        member: int,
        x_mag_A: float = 0,
        y_mag_A: float = 0,
        z_mag_A: float = 0,
        x_mag_B: float = 0,
        y_mag_B: float = 0,
        z_mag_B: float = 0,
        position_A: float = 0,
        position_B: float = 0,
        load_group: str = None,
        axes: str = 'global'
    ) -> int:
        """Create a distributed load with the next available ID.

        Args:
            member (int): Member to which the distributed load will be applied. Identified by the member ID.
            x_mag_A (float, optional): The magnitude of the load in the x-direction at the starting position A. Defaults to 0.
            y_mag_A (float, optional): The magnitude of the load in the y-direction at the starting position A. Defaults to 0.
            z_mag_A (float, optional): The magnitude of the load in the z-direction at the starting position A. Defaults to 0.
            x_mag_B (float, optional): The magnitude of the load in the x-direction at the finish position B. Defaults to 0.
            y_mag_B (float, optional): The magnitude of the load in the y-direction at the finish position B. Defaults to 0.
            z_mag_B (float, optional): The magnitude of the load in the z-direction at the finish position B. Defaults to 0.
            position_A (float, optional): Position along member where the distributed load starts. Expressed as a percentage. Defaults to 0.
            position_B (float, optional): Position along member where the distributed load ends. Expressed as a percentage. Defaults to 0.
            load_group (str, optional): The load group to which the load belongs. Defaults to None.
            axes (str, optional): The axes in which the distributed load will be applied. Defaults to 'global'.

        Returns:
            int: The ID of the created load.
        """

        next_index = next_object_key(self)

        dl = DistributedLoad(
            member,
            x_mag_A,
            y_mag_A,
            z_mag_A,
            x_mag_B,
            y_mag_B,
            z_mag_B,
            position_A,
            position_B,
            load_group,
            axes
        )
        setattr(self, str(next_index), dl)

        return next_index

    def id_from_member_id(self, element_id: int) -> int:
        """Find a distributed load's ID from the member ID which it is located.

        Args:
            element_id (int): The member ID of the distributed load to find.

        Returns:
            int: The distributed load ID which is located at the provided member ID.
        """
        found_id = None

        for k, v in vars(self).items():

            if (v.member == element_id):
                found_id = k

        return found_id



from typing import Literal


class Material:
    def __init__(
        self,
        name: str = None,
        density: float = None,
        elasticity_modulus: float = None,
        poissons_ratio: float = None,
        yield_strength: float = None,
        ultimate_strength: float = None,
        _class: Literal["steel", "aluminium",
                        "masonry", "concrete", "wood", "other"] = None
    ) -> None:
        """Creates an instance of the SkyCiv Material class.

        Args:
            name (str): The name of the material.
            density (float): The density of the material.
            elasticity_modulus (float): The Modulus of Elasticity of the material.
            poissons_ratio (float): The Poisson's Ratio for the material.
            yield_strength (float): The Yield strength of the material.
            ultimate_strength (float): The Ultimate strength the material.
            _class (str): The type of material. {"steel" | "aluminium" | "masonry" | "concrete" | "wood" | "other"}.
        """
        self.name = name
        self.density = density
        self.elasticity_modulus = elasticity_modulus
        self.poissons_ratio = poissons_ratio
        self.yield_strength = yield_strength
        self.ultimate_strength = ultimate_strength
        self.__setattr__("class", _class)  # Force with __setattr.



from skyciv.utils.helpers import clone


class OptionsObject:

    def __init__(self, validate_input: bool = True, response_data_only: bool = False) -> None:
        """Creates an instance of the SkyCiv OptionsObject class.

        Args:
            validate_input (bool, optional): Verify the model input with SkyCiv's built in model validator. Defaults to True.
            response_data_only (bool, optional): Only respond with data from the last function specified in the functions array. Defaults to False.
        """
        self.validate_input = validate_input
        self.response_data_only = response_data_only

    def get(self):
        return clone(self.__dict__)



from typing import Literal
from skyciv.classes.model.components.area_loads.area_loads import AreaLoads
from skyciv.classes.model.components.distributed_loads.distributed_loads import DistributedLoads
from skyciv.classes.model.components.load_combinations.load_combinations import LoadCombinations
from skyciv.classes.model.components.materials.materials import Materials
from skyciv.classes.model.components.members.members import Members
from skyciv.classes.model.components.meshed_plates.meshed_plates import MeshedPlates
from skyciv.classes.model.components.moments.moments import Moments
from skyciv.classes.model.components.nodes.nodes import Nodes
from skyciv.classes.model.components.plates.plates import Plates
from skyciv.classes.model.components.point_loads.point_loads import PointLoads
from skyciv.classes.model.components.pressures.pressures import Pressures
from skyciv.classes.model.components.sections.sections import Sections
from skyciv.classes.model.components.self_weights.self_weights import SelfWeights
from skyciv.classes.model.components.settings.settings import Settings
from skyciv.classes.model.components.settlements.settlements import Settlements
from skyciv.classes.model.components.supports.supports import Supports
from skyciv.utils.helpers import clone, has_get_method, keyvals


class Model:

    def __init__(self, unit_system: Literal["metric", "imperial"], vertical_axis: Literal["Y", "Z"] = 'Y'):
        """Create an s3d_model object.

        Args:
            unit_system (str): {'metric' | 'imperial'}
            vertical_axis (str, optional): {"Y" | "Z"}. Defaults to 'Y'.

        Example: 
            Default for: metric | imperial::
            {
                "length": "m" | 'ft',
                 "section_length": "mm" | 'in',
                 "material_strength": "mpa" | 'ksi',
                 "density": "kg/m3" | 'lb/ft3',
                 "force": "kn" | 'kip',
                 "moment": "kn-m" | 'kip-ft',
                 "pressure": "kpa" | 'ksf',
                 "mass": "kg" | 'kip',
                 "translation": "mm" | 'in',
                 "stress": "mpa" | 'ksi'
            }
        """
        self.settings = Settings(unit_system, vertical_axis)
        self.nodes = Nodes()
        self.members = Members()
        self.plates = Plates()
        self.meshed_plates = MeshedPlates()
        self.sections = Sections()
        self.materials = Materials()
        self.supports = Supports()
        self.settlements = Settlements()
        self.point_loads = PointLoads()
        self.moments = Moments()
        self.distributed_loads = DistributedLoads()
        self.pressures = Pressures()
        self.area_loads = AreaLoads()
        self.member_prestress_loads = {}
        self.self_weight = SelfWeights()
        self.load_combinations = LoadCombinations()
        self.load_cases = {}
        self.nodal_masses = {}
        self.nodal_masses_conversion_map = {}
        self.spectral_loads = {}
        self.groups = []

    def get(self) -> dict:
        """Get the model in the format required by the SkyCiv API.

        Returns:
            dict: A SkyCiv formatted S3D Model as a python dictionary.
        """
        s3d_model = clone(self.__dict__)

        for k, v in s3d_model.items():
            if (has_get_method(v)):
                s3d_model[k] = v.get()

        return s3d_model

    def set(self, model_object: dict) -> None:
        """Set individual properties of the model object.

        Args:
            model_object (dict): An object of key value pairs.

        Example::

            model = new Model()
            model.set({
                nodes: SkyCivNodesObject,
                members: "SkyCivMembersObject"
            })
        """
        for k, v in model_object.items():
            if hasattr(self, k):
                setattr(self, str(k), v)

    def __getitem__(self, item):
        return getattr(self, item)



from typing import Literal
from skyciv.utils.helpers import clone


class ApiFunction:

    def __init__(self, function_name: Literal["S3D.session.start", "S3D.model.set", "S3D.model.get", "S3D.model.repair", "S3D.model.solve", "S3D.model.takeScreenshot", "S3D.model.mesh", "S3D.results.set", "S3D.results.get", "S3D.results.fetchMemberResult", "S3D.results.getAnalysisReport", "S3D.model.takeScreenshot", "S3D.design.member.getInput", "S3D.design.member.check", "S3D.design.member.optimize", "standalone.member.check", "S3D.design.rc.getInput", "S3D.design.rc.check", "S3D.SB.loadLibraryShape", "S3D.SB.getLibraryTree", "S3D.SB.buildCustomShape", "S3D.SB.solve", "S3D.SB.runGSD", "S3D.file.save", "S3D.file.open", "S3D.file.share", "S3D.file.getFileDirectory"], args: dict = {}) -> None:
        """Creates an instance of the SkyCiv ApiFunction class. See the docs for available functions and respective arguments. https://skyciv.com/api/v3/docs/the-request-object#functions

        Args:
            function_name (str): The SkyCiv function name. {"S3D.session.start" | "S3D.model.set" | "S3D.model.get" | "S3D.model.repair" | "S3D.model.solve" | "S3D.model.takeScreenshot" | "S3D.model.mesh" | "S3D.results.set" | "S3D.results.get" | "S3D.results.fetchMemberResult" | "S3D.results.getAnalysisReport" | "S3D.model.takeScreenshot" | "S3D.design.member.getInput" | "S3D.design.member.check" | "S3D.design.member.optimize" | "standalone.member.check" | "S3D.design.rc.getInput" | "S3D.design.rc.check" | "S3D.SB.loadLibraryShape" | "S3D.SB.getLibraryTree" | "S3D.SB.buildCustomShape" | "S3D.SB.solve" | "S3D.SB.runGSD" | "S3D.file.save" | "S3D.file.open" | "S3D.file.share" | "S3D.file.getFileDirectory"}
            args (dict, optional): See the docs for the available arguments. Defaults to {}.
        """
        self.function = function_name
        self.arguments = args

    def get(self) -> dict:
        """Get the function object as a python dictionary.
        Returns:
            dict: The function as a python dictionary.
        """
        return clone(self.__dict__)



from skyciv.classes.model.components.settlements.settlement import Settlement
from skyciv.utils.helpers import keyvals, next_object_key
from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent


class Settlements(ModelCollectionComponent):
    """Creates an instance of the SkyCiv Settlements class.
    """

    def add(self,
            node: int,
            tx: float = 0,
            ty: float = 0,
            tz: float = 0,
            rx: float = 0,
            ry: float = 0,
            rz: float = 0
            ) -> int:
        """Create a settlement with the next available ID.

        Args:
            node (int): The ID of the node at which the settlement is applied.
            tx (float, optional): Settlement displacement in the global x-axis. Defaults to None.
            ty (float, optional): Settlement displacement in the global y-axis. Defaults to None.
            tz (float, optional): Settlement displacement in the global z-axis. Defaults to None.
            rx (float, optional): Settlement rotation about the global x-axis. Defaults to None.
            ry (float, optional): Settlement rotation about the global y-axis. Defaults to None.
            rz (float, optional): Settlement rotation about the global z-axis. Defaults to None.

        Returns:
            int: The ID of the created settlement.
        """
        next_index = next_object_key(self)

        settlement = Settlement(node, tx, ty, tz, rx, ry, rz)
        setattr(self, str(next_index), settlement)
        return next_index

    def id_from_node_id(self, node_id: int) -> int:
        """Find a settlement's ID from the node ID which it is located.

        Args:
            node_id (int): The node ID of the settlement to find.

        Returns:
            int: The ID of the located settlement.
        """
        id = None

        for k, v in keyvals(self):
            if (v.node == node_id):
                id = k

        return id



from typing import Literal


class DistributedLoad:

    def __init__(
        self,
        member: int = None,
        x_mag_A: int = 0,
        y_mag_A: int = 0,
        z_mag_A: int = 0,
        x_mag_B: int = 0,
        y_mag_B: int = 0,
        z_mag_B: int = 0,
        position_A: int = 0,
        position_B: int = 0,
        load_group: str = None,
        axes: Literal["global", "local"] = 'global'
    ) -> None:
        """Creates an instance of the SkyCiv DistributedLoad class.

        Args:
            member (int): Member to which the distributed load will be applied. Identified by the member ID.
            x_mag_A (int, optional): The magnitude of the load in the x-direction at the starting position A. Defaults to 0.
            y_mag_A (int, optional): The magnitude of the load in the y-direction at the starting position A. Defaults to 0.
            z_mag_A (int, optional): The magnitude of the load in the z-direction at the starting position A. Defaults to 0.
            x_mag_B (int, optional): The magnitude of the load in the x-direction at the finish position B. Defaults to 0.
            y_mag_B (int, optional): The magnitude of the load in the y-direction at the finish position B. Defaults to 0.
            z_mag_B (int, optional): The magnitude of the load in the z-direction at the finish position B. Defaults to 0.
            position_A (int, optional): Position along member where the distributed load starts. Expressed as a percentage. Defaults to 0.
            position_B (int, optional): Position along member where the distributed load ends. Expressed as a percentage. Defaults to 0.
            load_group (str, optional): The load group to which the load belongs. Defaults to None.
            axes (str, optional): The axes in which the distributed load will be applied. Defaults to 'global'. {"global" | "local"}.
        """
        self.member = member
        self.x_mag_A = x_mag_A
        self.y_mag_A = y_mag_A
        self.z_mag_A = z_mag_A
        self.x_mag_B = x_mag_B
        self.y_mag_B = y_mag_B
        self.z_mag_B = z_mag_B
        self.position_A = position_A
        self.position_B = position_B
        self.load_group = load_group
        self.axes = axes



import json
import http.client as httplib


def request(api_object: dict, options: dict = {}):
    """Makes a request to the SkyCiv API using the provided API object and options.

    Args:
        api_object (dict): Your SkyCiv API Object as a Python dictionary - see docs below for more information.
        options (dict): An optional dictionary with the following parameters
                 - http_or_https: string ("http" | "https"), defaults to "https"
                 - version: integer (2 | 3), defaults to 3

    Return:
        skyciv_response: The return value is a parsed Dictionary.

    Example::

        options = {"http_or_https": "https"}
        skyciv_response = skyciv.request(api_object, options)
        ...Do something with the response


    [SkyCiv API documentation](https://skyciv.com/api/v3/docs/the-request-object)

    """

    port = 8085
    version = 3

    if hasattr(options, "http_or_https"):
        # Validate options
        allowable_ports = ["http", "https"]
        if not any([x in options["http_or_https"] for x in allowable_ports]):
            raise Exception("Invalid argument passed to option http_or_https.")

        # Set port
        if (options["http_or_https"] == "http"):
            port = 8086

        # Set version
        if hasattr(options, "version"):
            version = options["version"]

    headers = {"Content-Type": "application/json"}
    conn = httplib.HTTPSConnection("api.skyciv.com", port)

    # Check if api_object has already been stringified, stringify it if not.
    stringified_data = api_object
    if not isinstance(stringified_data, str):
        stringified_data = json.dumps(api_object, separators=(",", ":"))

    conn.request("POST", "/v" + str(version), stringified_data, headers)

    raw_res = conn.getresponse()
    res_data = raw_res.read()
    conn.close()

    parsed_res = json.loads(res_data)

    if (parsed_res["response"]["status"] != 0):
        print(
            f'Unsuccessful solve with message: {parsed_res["response"]["msg"]}')

    return parsed_res



from typing import Literal
from skyciv.classes.model.components.settings.units import Units
from skyciv.utils.helpers import clone


class Settings:
    def __init__(self,
                 unit_system: Literal["metric", "imperial"],
                 vertical_axis: Literal["Y", "Z"] = 'Y'
                 ) -> None:
        """Create a settings object using the default values.

        Args:
            unit_system (str): {'metric' | 'imperial'}
            vertical_axis (str, optional): The global vertical axis. Defaults to 'Y'.
        """
        self.units = Units(unit_system)
        self.precision = 'fixed'
        self.precision_values = 3
        self.evaluation_points = 9
        self.vertical_axis = vertical_axis
        self.solver_timeout = 600
        self.accurate_buckling_shape = False
        self.buckling_johnson = False
        self.non_linear_tolerance = '1'
        self.non_linear_theory = 'small'
        self.auto_stabilize_model = False
        self.member_offsets_axis = 'global'

    def set(self, settings_object: dict) -> None:
        """Set individual properties of the settings object.

        Args:
            settings_object (dict): An object of key value pairs.

        Example::

            settings = SettingsObject()
            settings.set({
                "vertical_axis": "Z",
                "auto_stabilize_model": True
            })
        """
        for k, v in vars(settings_object).items():
            if self.hasattr(k):
                self[k] = v

    def get(self) -> None:
        """Get the settings object as a Python dictionary.
        """
        settings = clone(vars(self))
        settings["units"] = settings["units"].get()
        return settings

    def __getitem__(self, item):
        return getattr(self, item)



default_materials = {
    "metric": {
        "Structural Steel": {
            "name": "Structural Steel",
            "density": 7850,
            "elasticity_modulus": 200000,
            "poissons_ratio": 0.27,
            "yield_strength": 260,
            "ultimate_strength": 410,
            "class": "steel"
        },
        "Aluminium": {
            "name": "Aluminium",
            "density": 2700,
            "elasticity_modulus": 69000,
            "poissons_ratio": 0.32,
            "yield_strength": 100,
            "ultimate_strength": 150,
            "class": "aluminium"
        },
        "Carbon Fibre Reinforced Plastic": {
            "name": "Carbon Fibre Reinforced Plastic",
            "density": 3500,
            "elasticity_modulus": 150000,
            "poissons_ratio": 0.2,
            "yield_strength": None,
            "ultimate_strength": 2705,
            "class": "other"
        },
        "Concrete": {
            "name": "Concrete",
            "density": 2500,
            "elasticity_modulus": 17000,
            "poissons_ratio": 0.2,
            "yield_strength": None,
            "ultimate_strength": 3.5,
            "class": "concrete"
        },
        "Concrete High Strength": {
            "name": "Concrete High Strength",
            "density": 2500,
            "elasticity_modulus": 30000,
            "poissons_ratio": 0.2,
            "yield_strength": None,
            "ultimate_strength": 5,
            "class": "concrete"
        },
        "Oakwood": {
            "name": "Oakwood",
            "density": 900,
            "elasticity_modulus": 11000,
            "poissons_ratio": 0.3,
            "yield_strength": 4.5,
            "ultimate_strength": 5,
            "class": "wood"
        },
        "Glass": {
            "name": "Glass",
            "density": 2500,
            "elasticity_modulus": 70000,
            "poissons_ratio": 0.24,
            "yield_strength": None,
            "ultimate_strength": 33,
            "class": "other"
        }
    },
    "imperial": {
        "Structural Steel": {
            "name": "Structural Steel",
            "density": 490,
            "elasticity_modulus": 29000,
            "poissons_ratio": 0.27,
            "yield_strength": 36,
            "ultimate_strength": 58,
            "class": "steel"
        },
        "Aluminium": {
            "name": "Aluminium",
            "density": 167,
            "elasticity_modulus": 10000,
            "poissons_ratio": 0.32,
            "yield_strength": 15,
            "ultimate_strength": 22,
            "class": "aluminium"
        },
        "Carbon Fibre Reinforced Plastic": {
            "name": "Carbon Fibre Reinforced Plastic",
            "density": 200,
            "elasticity_modulus": 21800,
            "poissons_ratio": 0.2,
            "yield_strength": "",
            "ultimate_strength": 460,
            "class": "other"
        },
        "Concrete": {
            "name": "Concrete",
            "density": 150,
            "elasticity_modulus": 2500,
            "poissons_ratio": 0.2,
            "yield_strength": "",
            "ultimate_strength": 0.5,
            "class": "concrete"
        },
        "Concrete High Strength": {
            "name": "Concrete High Strength",
            "density": 150,
            "elasticity_modulus": 4400,
            "poissons_ratio": 0.2,
            "yield_strength": "",
            "ultimate_strength": 0.75,
            "class": "concrete"
        },
        "Oakwood": {
            "name": "Oakwood",
            "density": 56,
            "elasticity_modulus": 1600,
            "poissons_ratio": 0.3,
            "yield_strength": 0.65,
            "ultimate_strength": 0.75,
            "class": "wood"
        },
        "Glass": {
            "id": 7,
            "name": "Glass",
            "density": 160,
            "elasticity_modulus": 10000,
            "poissons_ratio": 0.24,
            "yield_strength": "",
            "ultimate_strength": 4.8,
            "class": "other"
        }
    }
}



class Support:
    def __init__(self,
                 node: int=None,
                 restraint_code: str = 'FFFFFF',
                 tx: float = 0,
                 ty: float = 0,
                 tz: float = 0,
                 rx: float = 0,
                 ry: float = 0,
                 rz: float = 0
                 ) -> None:
        """Creates an instance of the SkyCiv Support class.

        Args:
            node (int): The ID of the node at which the support is located.
            restraint_code (str, optional): A 6 character restraint code. See docs for restraint code https://skyciv.com/api/v3/docs/s3d-model#restraint-code. Defaults to 'FFFFFF'.
            tx (float, optional): Spring stiffness for translation in the x-axis. Applies only if the restraint code character for translation in x = S. Defaults to 0.
            ty (float, optional): Spring stiffness for translation in the y-axis. Applies only if the restraint code character for translation in y = S. Defaults to 0.
            tz (float, optional): Spring stiffness for translation in the z-axis. Applies only if the restraint code character for translation in z = S. Defaults to 0.
            rx (float, optional): Spring stiffness for rotation about the x-axis. Applies only if the restraint code character for rotation about x = S. Defaults to 0.
            ry (float, optional): Spring stiffness for rotation about the y-axis. Applies only if the restraint code character for rotation about y = S. Defaults to 0.
            rz (float, optional): Spring stiffness for rotation about the z-axis. Applies only if the restraint code character for rotation about z = S. Defaults to 0.
        """
        self.node = node

        if restraint_code:
            self.restraint_code = restraint_code
        else:
            self.restraint_code = "FFFFFF"

        self.tx = tx
        self.ty = ty
        self.tz = tz
        self.rx = rx
        self.ry = ry
        self.rz = rz



from typing import Literal
from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent
from skyciv.classes.model.components.materials.default_materials import default_materials
from skyciv.classes.model.components.materials.material import Material
from skyciv.utils.helpers import next_object_key


class Materials(ModelCollectionComponent):
    """Creates an instance of the SkyCiv Materials class.
    """

    def add_custom(
        self,
        name: str,
        density: float,
        elasticity_modulus: float,
        poissons_ratio: float,
        yield_strength: float,
        ultimate_strength: float,
        _class: Literal["steel", "aluminium",
                        "masonry", "concrete", "wood", "other"]
    ) -> int:
        """Create a custom material with the next available ID.

        Args:
            name (str): The name of the material.
            density (float): The density of the material.
            elasticity_modulus (float): The Modulus of Elasticity of the material.
            poissons_ratio (float): The Poisson's Ratio for the material.
            yield_strength (float): The Yield strength of the material.
            ultimate_strength (float): The Ultimate strength the material.
            _class (str): The type of material. {"steel" | "aluminium" | "masonry" | "concrete" | "wood" | "other"}.

        Returns:
            int: The ID of the created material.
        """

        next_index = next_object_key(self)

        material = Material(
            name,
            density,
            elasticity_modulus,
            poissons_ratio,
            yield_strength,
            ultimate_strength,
            _class
        )
        setattr(self, str(next_index), material)
        return next_index

    def add(self, material: Literal["Structural Steel",
                                    "Aluminium",
                                    "Carbon Fibre Reinforced Plastic",
                                    "Concrete",
                                    "Concrete High Strength",
                                    "Oakwood",
                                    "Glass"], unit_system: str = "metric") -> int:
        """Add a default material to the model.

        Args:
            material (str): The material name. {"Structural Steel" | "Aluminium" | "Carbon Fibre Reinforced Plastic" | "Concrete" | "Concrete High Strength" | "Oakwood" | "Glass"}.
            unit_system (str): {'metric' | 'imperial'}

        Returns:
            int: The ID of the created material.
        """

        materials = default_materials[unit_system][material]

        name = materials["name"]
        density = materials["density"]
        elasticity_modulus = materials["elasticity_modulus"]
        poissons_ratio = materials["poissons_ratio"]
        yield_strength = materials["yield_strength"]
        ultimate_strength = materials["ultimate_strength"]
        _class = materials["class"]

        index = self.add_custom(
            name,
            density,
            elasticity_modulus,
            poissons_ratio,
            yield_strength,
            ultimate_strength,
            _class
        )

        return index




from typing import Literal
from skyciv.utils.helpers import clone


class Units:
    def __init__(self, unit_system: Literal["metric", "imperial"], units_object: dict = None) -> None:
        """Create a units object using the default value for Metric or Imperial.

        Args:
            unit_system (str): {'metric' | 'imperial'}
            units_object (dict, optional): A dictionary of key, value pairs to set individual unit properties. Defaults to None.

        Example:
            Default values for metric | imperial::

            {
                "length": "m" | 'ft',
                "section_length": "mm" | 'in',
                "material_strength": "mpa" | 'ksi',
                "density": "kg/m3" | 'lb/ft3',
                "force": "kn" | 'kip',
                "moment": "kn-m" | 'kip-ft',
                "pressure": "kpa" | 'ksf',
                "mass": "kg" | 'kip',
                "translation": "mm" | 'in',
                "stress": "mpa" | 'ksi'
            }
        """
        self.length = None
        self.section_length = None
        self.material_strength = None
        self.density = None
        self.force = None
        self.moment = None
        self.pressure = None
        self.mass = None
        self.translation = None
        self.stress = None

        self.set_unit_system(unit_system)

        if (units_object):
            self.set(units_object)

    def set_unit_system(self, unit_system: Literal["metric", "imperial"]):
        """Set the units object using the default values for Metric or Imperial.

        Args:
            unit_system (str): {'metric' | 'imperial'}

        Example:
            Default values for metric | imperial::

            {
                "length": "m" | 'ft',
                "section_length": "mm" | 'in',
                "material_strength": "mpa" | 'ksi',
                "density": "kg/m3" | 'lb/ft3',
                "force": "kn" | 'kip',
                "moment": "kn-m" | 'kip-ft',
                "pressure": "kpa" | 'ksf',
                "mass": "kg" | 'kip',
                "translation": "mm" | 'in',
                "stress": "mpa" | 'ksi'
            }
        """
        if unit_system == 'metric':
            self.set_default_metric()
        else:
            self.set_default_imperial()

    def set_default_metric(self):
        """Set the unit system to the default metric values.

        Example::

            {
                "length": "m",
                "section_length": "mm",
                "material_strength": "mpa",
                "density": "kg/m3",
                "force": "kn",
                "moment": "kn-m",
                "pressure": "kpa",
                "mass": "kg",
                "translation": "mm",
                "stress": "mpa"
            }
        """
        self.length = 'm'
        self.section_length = 'mm'
        self.material_strength = 'mpa'
        self.density = 'kg/m3'
        self.force = 'kn'
        self.moment = 'kn-m'
        self.pressure = 'kpa'
        self.mass = 'kg'
        self.translation = 'mm'
        self.stress = 'mpa'

    def set_default_imperial(self):
        """Set the unit system to the default imperial values.

        Example::

            {
                length: 'ft',
                section_length: 'in',
                material_strength: 'ksi',
                density: 'lb/ft3',
                force: 'kip',
                moment: 'kip-ft',
                pressure: 'ksf',
                mass: 'kip',
                translation: 'in',
                stress: 'ksi',
            }
        """
        self.length = 'ft'
        self.section_length = 'in'
        self.material_strength = 'ksi'
        self.density = 'lb/ft3'
        self.force = 'kip'
        self.moment = 'kip-ft'
        self.pressure = 'ksf'
        self.mass = 'kip'
        self.translation = 'in'
        self.stress = 'ksi'

    def set(self, units_object: dict):
        """Set individual properties of the units object. Imperial and Metric MUST NOT be mixed.

        Args:
            units_object (dict): An object of key value pairs.

        Example::

            units_obj = Units("metric")
            units_obj.set({
                "length": "m",
                "pressure": "mpa"
            })
        """
        for k, v in vars(units_object).items():
            if (self.hasattr(k)):
                self[k] = v

    def get(self) -> None:
        """Get the units object as a Python dictionary.
        """
        return clone(vars(self))

    def get_unit_system(self) -> str:
        """Determines the unit system based on the length units.

        Returns:
            str: The unit system.
        """
        if (self.length in ['m', 'mm']):
            return 'metric'
        else:
            return 'imperial'

    def __getitem__(self, item):
        return getattr(self, item)



from typing import Literal


class Plate:
    def __init__(
        self,
        nodes: list[int] = None,
        thickness: float = None,
        material_id: int = None,
        rotZ: float = 0,
        type: str = 'auto',
        offset: float = 0,
        diaphragm: Literal["no", "rigid"] = 'no',
        membrane_thickness: float = None,
        shear_thickness: float = None,
        bending_thickness: float = None,
        state: Literal["stress", "strain"] = 'stress',
        holes: list[int] = None,
        is_meshed: bool = False
    ) -> None:
        """Creates an instance of the SkyCiv Plate class.

            Args:
                nodes (list[int], optional): An array of node IDs that define the plate. At least 3 IDs are required. Defaults to None.
                thickness (float, optional): Plate thickness. Defaults to None.
                material_id (int, optional): The ID of the material to use for the plate. Defaults to None.
                rotZ (float, optional): Rotation about the z-axis. Defaults to 0.
                type (str, optional): Auto will consider shear deformation when the plate thickness is sufficient. Defaults to 'auto'.
                offset (float, optional): Offset of the plate along its local z-axis. Defaults to 0.
                diaphragm (Literal[, optional): If the plate is a diaphragm.. Defaults to 'no'.
                membrane_thickness (float, optional): The membrane thickness to be used. Takes the value of thickness if None. Defaults to None.
                shear_thickness (float, optional): The shear thickness to be used. Takes the value of thickness if None. Defaults to None.
                bending_thickness (float, optional): The bending thickness to be used. Takes the value of thickness if None. Defaults to None.
                state (Literal[, optional): Denotes whether the plate is in a state of plane stress or plane strain. Defaults to 'stress'.
                holes (list[int], optional): An array of node IDs. Used to define holes in the plate.. Defaults to None.
                is_meshed (bool, optional): Indicates whether the plate is already meshed. Defaults to False.
        """
        self.nodes = nodes
        self.thickness = thickness
        self.material_id = material_id
        self.rotZ = rotZ
        self.type = type
        self.offset = offset
        self.diaphragm = diaphragm
        self.membrane_thickness = membrane_thickness
        self.shear_thickness = shear_thickness
        self.bending_thickness = bending_thickness
        self.state = state
        self.holes = holes
        self.is_meshed = is_meshed




class SelfWeight:
    def __init__(self,
                 enabled: bool = False,
                 x: float = 0,
                 y: float = 0,
                 z: float = 0,
                 LG: str = "SW1"
                 ) -> None:
        """Creates an instance of the SkyCiv SelfWeight class.

        Args:
            enabled (bool): If the this weight is applied to the model or not.
            x (float): Acceleration due to gravity in the x-axis, defined as a multiplier of the gravitational constant g.
            y (float): Acceleration due to gravity in the y-axis, defined as a multiplier of the gravitational constant g.
            z (float): Acceleration due to gravity in the z-axis, defined as a multiplier of the gravitational constant g.
            LG (str, optional): The load group that the self-weight belongs to. Defaults to LG1.
        """
        self.enabled = enabled
        self.x = x
        self.y = y
        self.z = z
        self.LG = LG

    def enable(self, x: float = 0, y: float = 0, z: float = 0, LG: str = "SW1") -> None:
        """Enables self-weight with provided gravity multipliers.

        Args:
            x (float, optional): Gravity multiplier in the x-direction. Defaults to 0.
            y (float, optional): Gravity multiplier in the y-direction. Defaults to 0.
            z (float, optional): Gravity multiplier in the z-direction. Defaults to 0.
            LG (str, optional): The load group that the self-weight belongs to. Defaults to LG1.

        Example:
            For a simple model where +Y on the global axis is vertical,
            the following will apply [1 ✕ gravity] for the model self-weight::

            sw = SelfWeight()
            sw.enable(0, -1, 0, "SW1")

        """
        self.enabled = True
        self.x = x
        self.y = y
        self.z = z
        self.LG = LG

    def disable(self) -> None:
        """Disables self-weight.
        """
        self.enabled = False



from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent
from skyciv.classes.model.components.load_combinations.load_combination import LoadCombination
from skyciv.utils.helpers import next_object_key


class LoadCombinations(ModelCollectionComponent):
    """Creates an instance of the SkyCiv LoadCombinations class.
    """

    def add(self, name: str, combination_factors: dict) -> int:
        """Adds a new load combination.

        Args:
            name (str): The name of the load combination.
            combination_factors (dict): Key value pairs for the factors to apply to the load groups.

        Returns:
            int: The ID of the created load combination.

        Example::

            lcs = LoadCombinations()

            factors = {
                "SW": 1,
                "windCase": 1,
                "liveLoad": 1.5
            }

            lcs.add("LC1", factors)
        """
        next_index = next_object_key(self)

        lc = LoadCombination(name, combination_factors)
        setattr(self, str(next_index), lc)

        return next_index



from typing import Literal
from skyciv.classes.model.components.members.member import Member
from skyciv.utils.helpers import next_object_key
from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent


class Members(ModelCollectionComponent):
    """Creates an instance of the SkyCiv Members class.
    """

    def add(
        self,
        node_A: int,
        node_B: int,
        section_id: int,
        fixity_A: str = 'FFFFFF',
        fixity_B: str = 'FFFFFF',
        type: Literal["normal", "normal_continuous",
                      "cable", "rigid"] = 'normal',
        cable_length: float = None
    ) -> int:
        """Create a member with the next available ID.

        Args:
            node_A (int): The ID of the node at the beginning of the member.
            node_B (int): The ID of the node at the end of the member.
            section_id (int): The ID of the section to apply to the member.
            fixity_A (str, optional): See docs for restraint code. https://skyciv.com/api/v3/docs/s3d-model#restraint-code. Defaults to 'FFFFFF'.
            fixity_B (str, optional): See docs for restraint code. https://skyciv.com/api/v3/docs/s3d-model#restraint-code. Defaults to 'FFFFFF'.
            type (str, optional): The type of member. Defaults to 'normal'. {"normal" | "normal_continuous" | "cable" | "rigid"}.
            cable_length (float, optional): Required only when type = cable. Defaults to None.

        Returns:
            int: Required only when type = cable.
        """
        next_index = next_object_key(self)
        element_ids = self.get_member_ids_from_nodes_ids(node_A, node_B)

        if (element_ids != None):
            print(
                'There is more than one member with the same end nodes. Ensure they have different offsets.')

        member = Member(
            node_A,
            node_B,
            section_id,
            fixity_A,
            fixity_B,
            type,
            cable_length
        )
        setattr(self, str(next_index), member)
        return next_index

    def get_member_ids_from_nodes_ids(self, node_A: int, node_B: int) -> list[int]:
        """Get the IDs of all members by its end nodes.

        Args:
            node_A (int): The ID of Node A.
            node_B (int): The ID of Node B.

        Returns:
            list[int]: An array of member IDs or None if none exist.
        """
        ids = []

        for k, v in vars(self).items():
            if (v.node_A == node_A and v.node_B == node_B):
                ids.append(k)

        if len(ids) == 0:
            ids = None

        return ids



from typing import Literal
from skyciv.classes.model.components.plates.plate import Plate
from skyciv.utils.helpers import next_object_key
from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent


class Plates(ModelCollectionComponent):
    """Creates an instance of the SkyCiv Plates class.
    """

    def add(self,
            nodes: list[int] = None,
            thickness: float = None,
            material_id: int = None,
            rotZ: float = 0,
            type: str = 'auto',
            offset: float = 0,
            diaphragm: Literal["no", "rigid"] = 'no',
            membrane_thickness: float = None,
            shear_thickness: float = None,
            bending_thickness: float = None,
            state: Literal["stress", "strain"] = 'stress',
            holes: list[int] = None,
            is_meshed: bool = False
            ) -> int:
        """Create a plate with the next available ID.

        Args:
            nodes (list[int], optional): An array of node IDs that define the plate. At least 3 IDs are required. Defaults to None.
            thickness (float, optional): Plate thickness. Defaults to None.
            material_id (int, optional): The ID of the material to use for the plate. Defaults to None.
            rotZ (float, optional): Rotation about the z-axis. Defaults to 0.
            type (str, optional): Auto will consider shear deformation when the plate thickness is sufficient. Defaults to 'auto'.
            offset (float, optional): Offset of the plate along its local z-axis. Defaults to 0.
            diaphragm (Literal[, optional): If the plate is a diaphragm.. Defaults to 'no'.
            membrane_thickness (float, optional): The membrane thickness to be used. Takes the value of thickness if None. Defaults to None.
            shear_thickness (float, optional): The shear thickness to be used. Takes the value of thickness if None. Defaults to None.
            bending_thickness (float, optional): The bending thickness to be used. Takes the value of thickness if None. Defaults to None.
            state (Literal[, optional): Denotes whether the plate is in a state of plane stress or plane strain. Defaults to 'stress'.
            holes (list[int], optional): An array of node IDs. Used to define holes in the plate.. Defaults to None.
            is_meshed (bool, optional): Indicates whether the plate is already meshed. Defaults to False.

        Returns:
            int: The ID of the created plate.
        """
        next_index = next_object_key(self)
        element_ids = self.get_plate_ids_from_node_ids(nodes)

        if (element_ids != None):
            print('There is more than one plate with the same nodes.')

        pl = Plate(nodes,
                   thickness,
                   material_id,
                   rotZ,
                   type,
                   offset,
                   diaphragm,
                   membrane_thickness,
                   shear_thickness,
                   bending_thickness,
                   state,
                   holes,
                   is_meshed,
                   )
        setattr(self, str(next_index), pl)

        return next_index

    def get_plate_ids_from_node_ids(self, nodes: list[int]) -> list[int]:
        """Get the IDs of all plates that match the provided nodes array. Node order IS considered.

        Args:
            nodes (list[int]): An array of node IDs that define the plate. At least 3 IDs are required.

        Returns:
            list[int]: An array of member IDs or None if none exist.
        """
        ids = []

        for k, v in vars(self).items():
            if v.nodes == nodes:
                ids.append(k)

        if len(ids) == 0:
            ids = None

        return ids



from skyciv.classes.model.components.meshed_plates.meshed_plate import MeshedPlate
from skyciv.utils.helpers import next_object_key
from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent


class MeshedPlates(ModelCollectionComponent):
    """Creates an instance of the SkyCiv MeshedPlates class.
    """

    def add(self, parent_plate: int, node_A: int, node_B: int, node_C: int, node_D: int = None, rotZ: float = 0) -> int:
        """Create a meshed plate with the next available ID.

        Args:
            parent_plate (int): The ID of the plate which this meshed plate originated from. Must refer to a plate in the plates object.
            node_A (int): The first node of the meshed plate.
            node_B (int): The second node of the meshed plate.
            node_C (int): The third node of the meshed plate.
            node_D (int, optional): The fourth node of the meshed plate. Set this to None if the meshed plate is triangular. Defaults to None.
            rotZ (float, optional): Rotation of this plate about the plate's local z-axis, in degrees. Defaults to 0.

        Returns:
            int: The ID of the new meshed plate element.
        """
        next_index = next_object_key(self)
        element_ids = self.get_meshed_plate_ids_from_nodes_ids(
            node_A, node_B, node_C, node_D)

        if (element_ids != None):
            print('There is more than one meshed plate with the same nodes.')

        mp = MeshedPlate(parent_plate, node_A, node_B, node_C, node_D, rotZ)
        setattr(self, str(next_index), mp)
        return next_index

    def get_meshed_plate_ids_from_nodes_ids(self, node_A: int, node_B: int, node_C: int, node_D: int = None) -> list[int]:
        """Get the IDs of all meshed plates by corner nodes.

        Args:
            node_A (int): The ID of Node A.
            node_B (int): The ID of Node B.
            node_C (int): The ID of Node C.
            node_D (int, optional): The ID of Node D. Defaults to None.

        Returns:
            list[int]: An array of meshed plate IDs or None if none exist.
        """
        ids = []

        for k, v in vars(self).items():
            if (
                v.node_A == node_A and
                v.node_B == node_B and
                v.node_C == node_C and
                v.node_D == node_D
            ):
                ids.append(k)

        if len(ids) == 0:
            ids = None

        return ids



from skyciv.utils.helpers import clone, keyvals


class ModelCollectionComponent:
    """The base class for model collection components.
    """

    def get(self, id: int = None) -> dict:
        """Get the data for a single element or all elements.

        Args:
            id (int, optional): The ID of the element to retrieve. Defaults to None.

        Returns:
            dict: All elements unless an element ID is provided.
        """
        data = {}

        if id is not None:
            if self.hasattr(id):
                setattr(data, id, clone(self[id]))
        else:

            # Trigger get on each single class
            for k, v in keyvals(self):
                data[k] = v.__dict__

        return data

    def exists(self, id: int) -> bool:
        """Check if an element exists.

        Args:
            id (int): The ID of the element to check for.

        Returns:
            bool: Whether the element exists.
        """
        if self.hasattr(id):
            return True
        else:
            return False

    def length(self) -> int:
        """Get the amount of elements in the collection.

        Returns:
            int: The amount of elements in the collection.
        """
        element_ids = self.__dict__.keys()
        return len(element_ids)

    def clear(self, id: int) -> None:
        """Remove an element.

        Args:
            id (int): The element ID to delete.
        """
        if self.hasattr(id):
            del self[id]

    def clear_all(self) -> None:
        """Remove all elements.
        """
        for key in self.__dict__.keys():
            del self[key]

    def __getitem__(self, item):
        return getattr(self, item)
